import enum
from abc import ABCMeta
from typing import Set


class ReservedTags(enum.Enum):
    ANNOTATION = "annotation"


# These are the base primitives that the data engine database is capable of storing
class MetadataFieldType(enum.Enum):
    BOOLEAN = "BOOLEAN"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    STRING = "STRING"
    BLOB = "BLOB"


# Inheritors of this ABC define custom types
# They are backed by a primitive type, but they also may have additional tags, describing specialized behavior
class DagshubDataType(metaclass=ABCMeta):
    """
    Attributes:
        backing_field_type: primitive type in the data engine database
        custom_tags: additional tags applied to this type
    """

    backing_field_type: MetadataFieldType = None
    custom_tags: Set[str] = None


class Int(DagshubDataType):
    backing_field_type = MetadataFieldType.INTEGER


class String(DagshubDataType):
    backing_field_type = MetadataFieldType.STRING


class Blob(DagshubDataType):
    backing_field_type = MetadataFieldType.BLOB


class Float(DagshubDataType):
    backing_field_type = MetadataFieldType.FLOAT


class Bool(DagshubDataType):
    backing_field_type = MetadataFieldType.BOOLEAN


class LabelStudioAnnotation(DagshubDataType):
    backing_field_type = MetadataFieldType.BLOB
    custom_tags = {ReservedTags.ANNOTATION.value}


class Voxel51Annotation(DagshubDataType):
    backing_field_type = MetadataFieldType.BLOB
    custom_tags = {ReservedTags.ANNOTATION.value}



import functools
from typing import Optional, Any, Dict, Union

from dagshub.data_engine.client.query_builder import GqlQuery


class GqlQueries:
    @staticmethod
    @functools.lru_cache()
    def datasource() -> str:
        q = (
            GqlQuery()
            .operation("query", name="datasource", input={"$id": "ID", "$name": "String"})
            .query(
                "datasource",
                input={
                    "id": "$id",
                    "name": "$name",
                },
            )
            .fields(
                [
                    "id",
                    "name",
                    "rootUrl",
                    "integrationStatus",
                    "preprocessingStatus",
                    "metadataFields {name valueType multiple tags}" "type",
                ]
            )
            .generate()
        )
        return q

    @staticmethod
    def datasource_params(id: Optional[Union[int, str]], name: Optional[str]) -> Dict[str, Any]:
        return {
            "id": id,
            "name": name,
        }

    @staticmethod
    @functools.lru_cache()
    def datasource_query(include_metadata: bool) -> str:
        metadata_fields = "metadata { key value }" if include_metadata else ""
        q = (
            GqlQuery()
            .operation(
                "query",
                name="datasourceQuery",
                input={
                    "$datasource": "ID!",
                    "$queryInput": "QueryInput",
                    "$first": "Int",
                    "$after": "String",
                },
            )
            .query(
                "datasourceQuery",
                input={
                    "datasource": "$datasource",
                    "filter": "$queryInput",
                    "first": "$first",
                    "after": "$after",
                },
            )
            .fields(
                [
                    f"edges {{ node {{ id path {metadata_fields} }} }}",
                    "pageInfo { hasNextPage endCursor }",
                ]
            )
            .generate()
        )
        return q

    @staticmethod
    def datasource_query_params(
        datasource_id: Union[int, str], query_input: Dict[str, Any], first: Optional[int], after: Optional[str]
    ) -> Dict[str, Any]:
        return {
            "datasource": datasource_id,
            "queryInput": query_input,
            "first": first,
            "after": after,
        }

    @staticmethod
    @functools.lru_cache()
    def dataset() -> str:
        q = (
            GqlQuery()
            .operation("query", name="dataset", input={"$id": "ID", "$name": "String"})
            .query(
                "dataset",
                input={
                    "dataset": "$id",
                    "name": "$name",
                },
            )
            .fields(
                [
                    "id",
                    "name",
                    "datasource {id name rootUrl integrationStatus preprocessingStatus "
                    "metadataFields {name valueType multiple tags} type}",
                    "datasetQuery",
                ]
            )
            .generate()
        )
        return q

    @staticmethod
    def dataset_params(id: Optional[Union[int, str]], name: Optional[str]) -> Dict[str, Any]:
        return {
            "id": id,
            "name": name,
        }



from fastai.learner import Learner, Recorder
from fastcore.basics import ignore_exceptions, patch, detuplify
from fastai.callback.core import Callback
from fastai.callback.hook import total_params
from fastai.torch_core import rank_distrib, to_detach

from ..logger import DAGsHubLogger as LoggerImpl


class DAGsHubLogger(Callback):
    """
    First, install the DAGsHub logger with
    `pip install dagshub`

    Example of Usage:
    ```
    from dagshub.fastai import DAGsHubLogger

    # To log only during a single training phase
    learn.fit(..., cbs=DAGsHubLogger())
    ```
    """

    "Saves model topology, losses & metrics"
    remove_on_fetch, order = True, Recorder.order + 1

    def __init__(
        self,
        metrics_path: str = "metrics.csv",
        should_log_metrics: bool = True,
        hparams_path: str = "params.yml",
        should_log_hparams: bool = True,
        should_make_dirs: bool = True,
    ):
        """
        :param metrics_path (str): Where to save the single metrics CSV file.
        :param should_log_metrics (bool): Whether to log metrics at all. Should probably always be True.
        :param hparams_path (str): Where to save the single hyperparameter YAML file.
        :param should_log_hparams (bool): Whether to log hyperparameters to a file.
            Should be False if you want to work with hyperparameters in a dependency file,
            rather than specifying them using command line arguments.
        :param should_make_dirs (bool): If true, the directory structure required by metrics_path
            and hparams_path will be created. Has no effect if the directory structure already exists.
        """
        super(DAGsHubLogger, self).__init__()
        self.logger = LoggerImpl(
            metrics_path=metrics_path,
            should_log_metrics=should_log_metrics,
            hparams_path=hparams_path,
            should_log_hparams=should_log_hparams,
            should_make_dirs=should_make_dirs,
            eager_logging=True,
        )
        self._dags_status_hyperparam_name = "success"
        self._dags_step_num = -1
        self._dags_epoch = 0

    def before_fit(self):
        # Make sure this is a training run
        self.run = not hasattr(self.learn, "lr_finder") and not hasattr(self, "gather_preds") and rank_distrib() == 0
        if not self.run:
            return

        # Log Hyper Parameters
        params = self.learn.gather_args()
        self.logger.log_hyperparams(params)

    def after_batch(self):
        if self.learn.training:
            self._dags_step_num += 1
            self._dags_epoch += 1 / self.n_iter
            hypers = {f"{k}_{i}": v for i, h in enumerate(self.opt.hypers) for k, v in h.items()}
            metrics = {
                "train_loss": to_detach(self.learn.smooth_loss.clone()).numpy(),
                "raw_loss": to_detach(self.learn.loss.clone()).numpy(),
                "epoch": self._dags_epoch,
                **hypers,
            }
            self.logger.log_metrics(metrics, step_num=self._dags_step_num)

    def after_epoch(self):
        self._dags_epoch = round(self._dags_epoch)
        metrics = {
            n: s
            for n, s in zip(self.recorder.metric_names, self.recorder.log)
            if n not in ["train_loss", "epoch", "time"]
        }
        metrics["epoch"] = self._dags_epoch
        self.logger.log_metrics(metrics, step_num=self._dags_step_num)

    def after_fit(self):
        self.run = True
        self._dags_epoch = round(self._dags_epoch)
        if (
            self._dags_status_hyperparam_name is not None
            and self._dags_status_hyperparam_name not in self.logger.hparams
        ):
            self.logger.log_hyperparams({self._dags_status_hyperparam_name: self.run})
        metrics = {
            n: s
            for n, s in zip(self.recorder.metric_names, self.recorder.log)
            if n not in ["train_loss", "epoch", "time"]
        }
        metrics["epoch"] = self._dags_epoch
        self.logger.log_metrics(metrics, step_num=self._dags_step_num)


@patch
def gather_args(self: Learner):
    """Gather config parameters accessible to the learner"""
    # args stored by `store_attr`
    cb_args = {f"{cb}": getattr(cb, "__stored_args__", True) for cb in self.cbs}
    args = {"Learner": self, **cb_args}
    # input dimensions
    try:
        n_inp = self.dls.train.n_inp
        args["n_inp"] = n_inp
        xb = self.dls.train.one_batch()[:n_inp]
        args.update(
            {f"input {n + 1} dim {i + 1}": d for n in range(n_inp) for i, d in enumerate(list(detuplify(xb[n]).shape))}
        )
    except Exception:
        print("Could not gather input dimensions")
    # other useful information
    with ignore_exceptions():
        args["batch size"] = self.dls.bs
        args["batch per epoch"] = len(self.dls.train)
        args["model parameters"] = total_params(self.model)[0]
        args["device"] = self.dls.device.type
        args["frozen"] = bool(self.opt.frozen_idx)
        args["frozen idx"] = self.opt.frozen_idx
        args["dataset.tfms"] = f"{self.dls.dataset.tfms}"
        args["dls.after_item"] = f"{self.dls.after_item}"
        args["dls.before_batch"] = f"{self.dls.before_batch}"
        args["dls.after_batch"] = f"{self.dls.after_batch}"
    return args



import configparser
import os
import urllib
from os.path import exists
from pathlib import Path

import git

from dagshub.auth import get_token
from dagshub.auth.token_auth import HTTPBearerAuth
from dagshub.common import config
from dagshub.common.helpers import get_project_root, http_request, log_message
from dagshub.upload import create_repo


def init(repo_name=None, repo_owner=None, url=None, root=None, host=config.host, mlflow=True, dvc=False):
    """
    Initialize a DagsHub repository.

    Initialization includes:
        1) Creates a repository on DagsHub if it doesn't exist yet
        2) If `dvc` flag is set, adds the DagsHub repository as a dvc remote
        3) If `mlflow` flag is set, initializes MLflow environment variables to enable logging experiments into the
            DagsHub hosted MLflow

    Arguments:
        root: path to the locally hosted git repository.
            If it's not set, tries to find a repository going up the folders
        repo_owner: along with `repo_name` defines the repository on DagsHub
        repo_name: along with `repo_owner` defines the repository on DagsHub
        url: url to the repository on DagsHub. Can be used as an alternative to repo_owner/repo_name arguments
        host: address of a hosted DagsHub instance
        mlflow: configure MLflow to log experiments to DagsHub
        dvc: configure a dvc remote in the repository
    """
    # Setup required variables
    if dvc:
        root = root or get_project_root(Path(os.path.abspath(".")))
        if not exists(root / ".git"):
            raise ValueError(
                f"No git project found! (stopped at mountpoint {root}). \
                               Please run this command in a git repository."
            )

    if url and (repo_name or repo_owner):
        repo_name, repo_owner = None, None

    if not url:
        if repo_name is not None and repo_owner is not None:
            url = urllib.parse.urljoin(f"{host}/", f"{repo_owner}/{repo_name}")
        elif dvc:
            for remote in git.Repo(root).remotes:
                if host in remote.url:
                    url = remote.url[:-4]
    if not url:
        raise ValueError("No host remote found! Please specify the remote using the url variable, or --url argument.")
    elif url[-4] == ".":
        url = url[:-4]

    if not (repo_name and repo_owner):
        splitter = lambda x: (x[-1], x[-2])  # noqa E721
        repo_name, repo_owner = splitter(url.split("/"))

    if None in [repo_name, repo_owner, url]:
        raise ValueError(
            "Could not parse repository owner and name. Make sure you specify either a link \
                          to the repository with --url or a pair of --repo-owner and --repo-name"
        )

    # Setup authentication
    token = config.token or get_token(host=host)
    bearer = HTTPBearerAuth(token)

    # Configure repository
    res = http_request(
        "GET",
        urllib.parse.urljoin(f"{host}/", config.REPO_INFO_URL.format(owner=repo_owner, reponame=repo_name)),
        auth=bearer,
    )
    if res.status_code == 404:
        create_repo(repo_name)

    # Configure MLFlow
    if mlflow:
        os.environ["MLFLOW_TRACKING_URI"] = f"{url}.mlflow"
        os.environ["MLFLOW_TRACKING_USERNAME"] = token
        os.environ["MLFLOW_TRACKING_PASSWORD"] = token

    # Configure DVC
    if dvc:
        Path(root / ".dvc").mkdir(parents=True, exist_ok=True)
        write = True

        dvc_config = configparser.ConfigParser()
        dvc_config_local = configparser.ConfigParser()
        dvc_config.read(root / ".dvc" / "config")
        dvc_config_local.read(root / ".dvc" / "config.local")

        for section in dvc_config.sections():
            if "url" in dvc_config[section] and host in dvc_config[section]["url"]:
                write = False

        remote = "dagshub" if "origin" in dvc_config.sections() else "origin"
        if write:
            dvc_config_local[f"'remote \"{remote}\"'"] = {"auth": "basic", "user": token, "password": token}
            dvc_config[f"'remote \"{remote}\"'"] = {"url": f"{url}.dvc"}

            with open(root / ".dvc" / "config", "w") as config_file, open(
                root / ".dvc" / "config.local", "w"
            ) as config_local_file:
                dvc_config.write(config_file)
                dvc_config_local.write(config_local_file)
                log_message(f'Added new remote "{remote}" with url = {url}')

        if not exists(root / ".dvc" / ".gitignore"):
            with open(root / ".dvc" / ".gitignore", "w") as config_gitignore:
                config_gitignore.write(config.CONFIG_GITIGNORE)

    log_message("Repository initialized!")



import functools
from typing import Any, Dict, List, Union, Optional

from dagshub.data_engine.client.query_builder import GqlQuery

from dagshub.data_engine.client.models import DatasourceType, ScanOption


class GqlMutations:
    @staticmethod
    @functools.lru_cache()
    def create_datasource():
        q = (
            GqlQuery()
            .operation(
                "mutation",
                name="createDatasource",
                input={"$name": "String!", "$url": "String!", "$dsType": "DatasourceType!"},
            )
            .query("createDatasource", input={"name": "$name", "url": "$url", "dsType": "$dsType"})
            .fields(["id", "name", "rootUrl", "integrationStatus", "preprocessingStatus", "type"])
            .generate()
        )
        return q

    @staticmethod
    def create_datasource_params(name: str, url: str, ds_type: DatasourceType):
        return {"name": name, "url": url, "dsType": str(ds_type.value)}

    @staticmethod
    @functools.lru_cache()
    def update_metadata():
        q = (
            GqlQuery()
            .operation(
                "mutation",
                name="updateMetadata",
                input={"$datasource": "ID!", "$datapoints": "[DatapointMetadataInput!]!"},
            )
            .query("updateMetadata", input={"datasource": "$datasource", "datapoints": "$datapoints"})
            .fields(
                [
                    "path",
                ]
            )
            .generate()
        )
        return q

    @staticmethod
    @functools.lru_cache()
    def update_metadata_field():
        q = (
            GqlQuery()
            .operation(
                "mutation",
                name="updateMetadataFieldProps",
                input={"$datasource": "ID!", "$props": "[MetadataFieldPropsInput!]!"},
            )
            .query("updateMetadataFieldProps", input={"datasource": "$datasource", "props": "$props"})
            .fields(
                [
                    "name",
                    "valueType",
                    "multiple",
                    "tags",
                ]
            )
            .generate()
        )
        return q

    @staticmethod
    def update_metadata_params(datasource_id: Union[int, str], datapoints: List[Dict[str, Any]]):
        return {
            "datasource": datasource_id,
            "datapoints": datapoints,
        }

    def update_metadata_fields_params(datasource_id: Union[int, str], metadata_field_props: List[Dict[str, Any]]):
        return {"datasource": datasource_id, "props": metadata_field_props}

    @staticmethod
    @functools.lru_cache()
    def delete_datasource():
        q = (
            GqlQuery()
            .operation(
                "mutation",
                name="deleteDatasource",
                input={
                    "$id": "ID!",
                },
            )
            .query(
                "deleteDatasource",
                input={
                    "id": "$id",
                },
            )
            .fields(
                [
                    "id",
                    "name",
                    "rootUrl",
                    "integrationStatus",
                    "preprocessingStatus",
                    "type",
                ]
            )
            .generate()
        )
        return q

    @staticmethod
    def delete_datasource_params(datasource_id: Union[int, str]):
        return {
            "id": datasource_id,
        }

    @staticmethod
    @functools.lru_cache()
    def scan_datasource():
        q = (
            GqlQuery()
            .operation(
                "mutation",
                name="scanDatasource",
                input={
                    "$id": "ID!",
                    "$options": "[ScanOption]",
                },
            )
            .query(
                "scanDatasource",
                input={
                    "id": "$id",
                    "options": "$options",
                },
            )
            .fields(
                [
                    "id",
                    "name",
                    "rootUrl",
                    "integrationStatus",
                    "preprocessingStatus",
                    "type",
                ]
            )
            .generate()
        )
        return q

    @staticmethod
    def scan_datasource_params(datasource_id: Union[int, str], options: Optional[List[ScanOption]]):
        return {
            "id": datasource_id,
            "options": options,
        }

    @staticmethod
    @functools.lru_cache()
    def save_dataset():
        q = (
            GqlQuery()
            .operation(
                "mutation",
                name="saveDataset",
                input={"$datasource": "ID!", "$name": "String!", "$filter": "QueryInput!"},
            )
            .query("saveDataset", input={"datasource": "$datasource", "name": "$name", "filter": "$filter"})
            .fields(
                [
                    "id",
                    "name",
                    "createdAt",
                ]
            )
            .generate()
        )
        return q

    @staticmethod
    def save_dataset_params(datasource_id: Union[int, str], name: str, query_input: Dict[str, Any]):
        return {
            "datasource": datasource_id,
            "name": name,
            "filter": query_input,
        }
